<?php

namespace Drupal\modeler_api\Hook;

use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Extension\Extension;
use Drupal\Core\Hook\Attribute\Hook;
use Drupal\Core\Routing\RouteBuilderInterface;
use Drupal\Core\Routing\RouteProviderInterface;
use Drupal\Core\Url;
use Drupal\modeler_api\Api;
use Drupal\modeler_api\Entity\AccessControlHandler;
use Drupal\modeler_api\Entity\ListBuilder;
use Drupal\modeler_api\Form\DeleteForm;
use Drupal\modeler_api\Plugin\ModelerPluginManager;
use Drupal\modeler_api\Plugin\ModelOwnerPluginManager;
use Symfony\Component\Routing\Exception\RouteNotFoundException;

/**
 * Implements hooks for the Modeler API module.
 */
class EntityHooks {

  /**
   * Constructs a new ModelerApiHooks object.
   *
   * @param \Drupal\modeler_api\Api $modelerApiService
   *   The modeler API service.
   * @param \Drupal\modeler_api\Plugin\ModelerPluginManager $modelerManager
   *   The modeler plugin manager.
   * @param \Drupal\modeler_api\Plugin\ModelOwnerPluginManager $modelOwnerPluginManager
   *   The model owner plugin manager.
   * @param \Drupal\Core\Routing\RouteProviderInterface $routeProvider
   *   The route provider.
   * @param \Drupal\Core\Routing\RouteBuilderInterface $routeBuilder
   *   The route builder.
   */
  public function __construct(
    protected Api $modelerApiService,
    protected ModelerPluginManager $modelerManager,
    protected ModelOwnerPluginManager $modelOwnerPluginManager,
    protected RouteProviderInterface $routeProvider,
    protected RouteBuilderInterface $routeBuilder,
  ) {}

  /**
   * Implements hook_entity_type_build().
   *
   * @param array $entity_types
   *   The entity type definitions.
   */
  #[Hook('entity_type_build')]
  public function entityTypeBuild(array &$entity_types): void {
    static $alreadyRunning;
    if ($alreadyRunning) {
      return;
    }
    $alreadyRunning = TRUE;
    foreach ($this->modelOwnerPluginManager->getAllInstances() as $owner) {
      $basePath = $owner->configEntityBasePath();
      if ($basePath === NULL) {
        continue;
      }
      $type = $owner->configEntityTypeId();
      /**
       * @var \Drupal\Core\Entity\EntityTypeInterface[] $entity_types
       */
      $entity_types[$type]
        ->setAccessClass(AccessControlHandler::class)
        ->setListBuilderClass(ListBuilder::class)
        ->setLinkTemplate('collection', '/' . $basePath)
        ->setFormClass('delete', DeleteForm::class)
        ->setLinkTemplate('edit-form', '/' . $basePath . '/{' . $type . '}/edit')
        ->setLinkTemplate('delete-form', '/' . $basePath . '/{' . $type . '}/delete');
    }
    $alreadyRunning = FALSE;
  }

  /**
   * Implements hook_entity_operation().
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @return array
   *   An array of operations.
   */
  #[Hook('entity_operation')]
  public function entityOperation(EntityInterface $entity): array {
    $operations = [];
    if (
      $entity instanceof ConfigEntityInterface &&
      ($owner = $this->modelerApiService->findOwner($entity)) &&
      $owner->configEntityBasePath()
    ) {
      $type = $entity->getEntityTypeId();
      $modelerId = $owner->getModelerId($entity);
      foreach ($this->modelerManager->getAllInstances() as $id => $modeler) {
        if ($modelerId !== $id || $owner->getModelData($entity) === '') {
          if ($modeler->isEditable()) {
            $operations['open_with_' . $id] = [
              'title' => t('Edit with :label', [':label' => $modeler->label()]),
              'url' => Url::fromRoute('entity.' . $type . '.edit_with.' . $id, [$type => $entity->id()]),
              'weight' => 49,
            ];
          }
        }
      }
      if (!$entity->status()) {
        $operations['enable'] = [
          'title' => t('Enable'),
          'url' => Url::fromRoute('entity.' . $type . '.enable', [$type => $entity->id()]),
          'weight' => 50,
        ];
      }
      else {
        $operations['disable'] = [
          'title' => t('Disable'),
          'url' => Url::fromRoute('entity.' . $type . '.disable', [$type => $entity->id()]),
          'weight' => 51,
        ];
      }
      if ($owner->isEditable($entity)) {
        $operations['clone'] = [
          'title' => t('Clone'),
          'url' => Url::fromRoute('entity.' . $type . '.clone', [$type => $entity->id()]),
          'weight' => 52,
        ];
      }
      if ($owner->isExportable($entity)) {
        $operations['export'] = [
          'title' => t('Export'),
          'url' => Url::fromRoute('entity.' . $type . '.export', [$type => $entity->id()]),
          'weight' => 52,
        ];
        $operations['export_recipe'] = [
          'title' => t('Export as recipe'),
          'url' => Url::fromRoute('entity.' . $type . '.export_recipe', [$type => $entity->id()]),
          'weight' => 53,
        ];
      }
    }
    return $operations;
  }

  /**
   * Implements hook_config_schema_info_alter().
   *
   * @param array $definitions
   *   The schema definitions.
   */
  #[Hook('config_schema_info_alter')]
  public function configSchemaInfoAlter(array &$definitions): void {
    foreach ($this->modelOwnerPluginManager->getAllInstances() as $owner) {
      $provider = $owner->configEntityProviderId();
      $entityTypeId = $owner->configEntityTypeId();
      $key = implode('.', [$provider, $entityTypeId, '*']);
      if (!isset($definitions[$key]['mapping']['third_party_settings']['mapping']['modeler_api'])) {
        $definitions[$key]['mapping']['third_party_settings']['type'] = 'mapping';
        $definitions[$key]['mapping']['third_party_settings']['mapping'] = [
          'modeler_api' => [
            'type' => 'mapping',
            'mapping' => [
              'modeler_id' => [
                'type' => 'string',
                'label' => 'ID',
              ],
              'data' => [
                'type' => 'string',
                'label' => 'Raw data, or an md5 hash of the raw data if that is stored externally',
              ],
              'changelog' => [
                'type' => 'label',
                'label' => 'Changelog',
              ],
              'label' => [
                'type' => 'label',
                'label' => 'Label',
              ],
              'documentation' => [
                'type' => 'label',
                'label' => 'Documentation',
              ],
              'tags' => [
                'type' => 'mapping',
                'label' => 'Changelog',
              ],
              'version' => [
                'type' => 'string',
                'label' => 'Version',
              ],
            ],
          ],
        ];
      }
    }
  }

  /**
   * Implements hook_system_info_alter().
   *
   * @param array $info
   *   The info file contents, passed by reference so that it can be altered.
   * @param \Drupal\Core\Extension\Extension $file
   *   Full information about the module or theme.
   * @param string $type
   *   Either 'module' or 'theme', depending on the type of .info.yml file that
   *   was passed.
   */
  #[Hook('system_info_alter')]
  public function systemInfoAlter(array &$info, Extension $file, string $type): void {
    if ($type === 'module') {
      foreach ($this->modelOwnerPluginManager->getAllInstances() as $owner) {
        $provider = $owner->getPluginDefinition()['provider'];
        if (str_ends_with($file->getPathname(), '/' . $provider . '.info.yml')) {
          $route = 'entity.' . $owner->configEntityTypeId() . '.collection';
          try {
            $this->routeProvider->getRouteByName($route);
            $info['configure'] = $route;
          }
          catch (RouteNotFoundException) {
            // If the route can not be found, don't set the configure route.
          }
        }
      }
    }
  }

}
